@@ -11,6 +11,9 @@
 #include "distmeta.h"
 #include "_amqstate.h"
 
+#define PYRABBITMQ_CONNECTION_ERROR 0x10
+#define PYRABBITMQ_CHANNEL_ERROR 0x20
+
 /* ------: Private Prototypes :------------------------------------------- */
 PyMODINIT_FUNC init_librabbitmq(void);
 
@@ -25,14 +28,35 @@ PyObject *PyRabbitMQ_socket_timeout;
 _PYRMQ_INLINE amqp_table_entry_t*
 AMQTable_AddEntry(amqp_table_t*, amqp_bytes_t);
 
+_PYRMQ_INLINE void
+AMQTable_SetTableValue(amqp_table_t*, amqp_bytes_t, amqp_table_t);
+
+_PYRMQ_INLINE void
+AMQTable_SetArrayValue(amqp_table_t*, amqp_bytes_t, amqp_array_t);
+
 _PYRMQ_INLINE void
 AMQTable_SetStringValue(amqp_table_t*, amqp_bytes_t, amqp_bytes_t);
 
 _PYRMQ_INLINE void
 AMQTable_SetIntValue(amqp_table_t *, amqp_bytes_t, int);
 
-_PYRMQ_INLINE amqp_table_t
-PyDict_ToAMQTable(amqp_connection_state_t, PyObject *);
+_PYRMQ_INLINE void
+AMQTable_SetDoubleValue(amqp_table_t *, amqp_bytes_t, double);
+
+_PYRMQ_INLINE amqp_field_value_t*
+AMQArray_AddEntry(amqp_array_t*);
+
+_PYRMQ_INLINE void
+AMQArray_SetTableValue(amqp_array_t*, amqp_table_t);
+
+_PYRMQ_INLINE void
+AMQArray_SetArrayValue(amqp_array_t*, amqp_array_t);
+
+_PYRMQ_INLINE void
+AMQArray_SetStringValue(amqp_array_t*, amqp_bytes_t);
+
+_PYRMQ_INLINE void
+AMQArray_SetIntValue(amqp_array_t *, int);
 
 _PYRMQ_INLINE int64_t RabbitMQ_now_usec(void);
 _PYRMQ_INLINE int RabbitMQ_wait_nb(int);
@@ -59,12 +83,24 @@ int
 PyRabbitMQ_recv(PyRabbitMQ_Connection *, PyObject *,
                 amqp_connection_state_t, int);
 
+unsigned int
+PyRabbitMQ_revive_channel(PyRabbitMQ_Connection *, unsigned int);
+
+    unsigned int
+PyRabbitMQ_Connection_create_channel(PyRabbitMQ_Connection *, unsigned int);
+
 int PyRabbitMQ_HandleError(int, char const *);
 _PYRMQ_INLINE int PyRabbitMQ_HandlePollError(int);
-int PyRabbitMQ_HandleAMQError(amqp_rpc_reply_t, PyObject *, const char *);
+int PyRabbitMQ_HandleAMQError(PyRabbitMQ_Connection *, unsigned int,
+                              amqp_rpc_reply_t, const char *);
 void PyRabbitMQ_SetErr_UnexpectedHeader(amqp_frame_t*);
 int PyRabbitMQ_Not_Connected(PyRabbitMQ_Connection *);
 
+static amqp_table_t PyDict_ToAMQTable(amqp_connection_state_t, PyObject *);
+static amqp_array_t PyList_ToAMQArray(amqp_connection_state_t, PyObject *);
+
+static PyObject* AMQTable_toPyDict(amqp_table_t *table);
+static PyObject* AMQArray_toPyList(amqp_array_t *array);
 
 
 int PyRabbitMQ_Not_Connected(PyRabbitMQ_Connection *self)
@@ -88,6 +124,24 @@ AMQTable_AddEntry(amqp_table_t *table, amqp_bytes_t key)
     return entry;
 }
 
+_PYRMQ_INLINE void
+AMQTable_SetTableValue(amqp_table_t *table,
+                       amqp_bytes_t key, amqp_table_t value)
+{
+    amqp_table_entry_t *entry = AMQTable_AddEntry(table, key);
+    entry->value.kind = AMQP_FIELD_KIND_TABLE;
+    entry->value.value.table = value;
+}
+
+_PYRMQ_INLINE void
+AMQTable_SetArrayValue(amqp_table_t *table,
+                       amqp_bytes_t key, amqp_array_t value)
+{
+    amqp_table_entry_t *entry = AMQTable_AddEntry(table, key);
+    entry->value.kind = AMQP_FIELD_KIND_ARRAY;
+    entry->value.value.array = value;
+}
+
 _PYRMQ_INLINE void
 AMQTable_SetStringValue(amqp_table_t *table,
                         amqp_bytes_t key, amqp_bytes_t value)
@@ -106,7 +160,56 @@ AMQTable_SetIntValue(amqp_table_t *table,
     entry->value.value.i32 = value;
 }
 
-_PYRMQ_INLINE amqp_table_t
+_PYRMQ_INLINE void
+AMQTable_SetDoubleValue(amqp_table_t *table,
+                        amqp_bytes_t key, double value)
+{
+    amqp_table_entry_t *entry = AMQTable_AddEntry(table, key);
+    entry->value.kind = AMQP_FIELD_KIND_F64;
+    entry->value.value.f64 = value;
+}
+
+_PYRMQ_INLINE amqp_field_value_t*
+AMQArray_AddEntry(amqp_array_t *array)
+{
+    amqp_field_value_t *entry = &array->entries[array->num_entries];
+    array->num_entries++;
+    return entry;
+}
+
+_PYRMQ_INLINE void
+AMQArray_SetTableValue(amqp_array_t *array, amqp_table_t value)
+{
+    amqp_field_value_t *entry = AMQArray_AddEntry(array);
+    entry->kind = AMQP_FIELD_KIND_TABLE;
+    entry->value.table = value;
+}
+
+_PYRMQ_INLINE void
+AMQArray_SetArrayValue(amqp_array_t *array, amqp_array_t value)
+{
+    amqp_field_value_t *entry = AMQArray_AddEntry(array);
+    entry->kind = AMQP_FIELD_KIND_ARRAY;
+    entry->value.array = value;
+}
+
+_PYRMQ_INLINE void
+AMQArray_SetStringValue(amqp_array_t *array, amqp_bytes_t value)
+{
+    amqp_field_value_t *entry = AMQArray_AddEntry(array);
+    entry->kind = AMQP_FIELD_KIND_UTF8;
+    entry->value.bytes = value;
+}
+
+_PYRMQ_INLINE void
+AMQArray_SetIntValue(amqp_array_t *array, int value)
+{
+    amqp_field_value_t *entry = AMQArray_AddEntry(array);
+    entry->kind = AMQP_FIELD_KIND_I32;
+    entry->value.i32 = value;
+}
+
+static amqp_table_t
 PyDict_ToAMQTable(amqp_connection_state_t conn, PyObject *src)
 {
     PyObject *dkey = NULL;
@@ -114,6 +217,7 @@ PyDict_ToAMQTable(amqp_connection_state_t conn, PyObject *src)
     PY_SIZE_TYPE size = 0;
     PY_SIZE_TYPE pos = 0;
     uint64_t clong_value = 0;
+    double cdouble_value = 0.0;
     int is_unicode = 0;
     amqp_table_t dst = AMQP_EMPTY_TABLE;
 
@@ -125,11 +229,27 @@ PyDict_ToAMQTable(amqp_connection_state_t conn, PyObject *src)
                            size * sizeof(amqp_table_entry_t));
     while (PyDict_Next(src, &pos, &dkey, &dvalue)) {
 
-        /* Int | Long */
-        if (PyLong_Check(dvalue) || PyInt_Check(dvalue)) {
+        if (PyDict_Check(dvalue)) {
+            /* Dict */
+            AMQTable_SetTableValue(&dst,
+                    PyString_AS_AMQBYTES(dkey),
+                    PyDict_ToAMQTable(conn, dvalue));
+        }
+        else if (PyList_Check(dvalue)) {
+            /* List */
+            AMQTable_SetArrayValue(&dst,
+                    PyString_AS_AMQBYTES(dkey),
+                    PyList_ToAMQArray(conn, dvalue));
+        }
+        else if (PyLong_Check(dvalue) || PyInt_Check(dvalue)) {
+            /* Int | Long */
             clong_value = (int64_t)PyLong_AsLong(dvalue);
+            if (PyErr_Occurred())
+                goto error;
             AMQTable_SetIntValue(&dst,
-                    PyString_AS_AMQBYTES(dkey), clong_value);
+                    PyString_AS_AMQBYTES(dkey),
+                    clong_value
+            );
         }
         else {
             /* String | Unicode */
@@ -141,14 +261,27 @@ PyDict_ToAMQTable(amqp_connection_state_t conn, PyObject *src)
                 }
                 AMQTable_SetStringValue(&dst,
                         PyString_AS_AMQBYTES(dkey),
-                        PyString_AS_AMQBYTES(dvalue));
+                        PyString_AS_AMQBYTES(dvalue)
+                );
             }
             else {
-                /* unsupported type */
-                PyErr_Format(PyExc_ValueError,
-                    "Table member %s is of an unsupported type",
-                    PyString_AS_STRING(dkey));
-                goto error;
+                cdouble_value = PyFloat_AsDouble(dvalue);
+                if (PyErr_Occurred())
+                    goto error;
+                if (PyFloat_Check(dvalue)) {
+
+                    AMQTable_SetDoubleValue(&dst,
+                        PyString_AS_AMQBYTES(dkey),
+                        cdouble_value
+                    );
+                }
+                else {
+                    /* unsupported type */
+                    PyErr_Format(PyExc_ValueError,
+                        "Table member %s is of an unsupported type",
+                        PyString_AS_STRING(dkey));
+                    goto error;
+            }
             }
         }
     }
@@ -158,6 +291,67 @@ error:
     return AMQP_EMPTY_TABLE;
 }
 
+static amqp_array_t
+PyList_ToAMQArray(amqp_connection_state_t conn, PyObject *src)
+{
+    PyObject *dvalue = NULL;
+    PY_SIZE_TYPE size = 0;
+    PY_SIZE_TYPE pos = 0;
+    uint64_t clong_value = 0;
+    int is_unicode = 0;
+    amqp_array_t dst = AMQP_EMPTY_ARRAY;
+
+    size = PyList_Size(src);
+
+    /* allocate new array */
+    dst.num_entries = 0;
+    dst.entries = amqp_pool_alloc(&conn->frame_pool,
+                           size * sizeof(amqp_field_value_t));
+    for (pos = 0; pos < size; ++pos) {
+
+        dvalue = PyList_GetItem(src, pos);
+
+        if (PyDict_Check(dvalue)) {
+            /* Dict */
+            AMQArray_SetTableValue(&dst,
+                    PyDict_ToAMQTable(conn, dvalue));
+        }
+        else if (PyList_Check(dvalue)) {
+            /* List */
+            AMQArray_SetArrayValue(&dst,
+                    PyList_ToAMQArray(conn, dvalue));
+        }
+        else if (PyLong_Check(dvalue) || PyInt_Check(dvalue)) {
+            /* Int | Long */
+            clong_value = (int64_t)PyLong_AsLong(dvalue);
+            AMQArray_SetIntValue(&dst, clong_value);
+        }
+        else {
+            /* String | Unicode */
+            is_unicode = PyUnicode_Check(dvalue);
+            if (is_unicode || PyString_Check(dvalue)) {
+                if (is_unicode) {
+                    if ((dvalue = PyUnicode_AsASCIIString(dvalue)) == NULL)
+                        goto error;
+                }
+                AMQArray_SetStringValue(&dst,
+                        PyString_AS_AMQBYTES(dvalue));
+            }
+            else {
+                /* unsupported type */
+                PyErr_Format(PyExc_ValueError,
+                    "Array member at index %lu, %s, is of an unsupported type",
+                    pos, PyObject_REPR(dvalue));
+                goto error;
+            }
+        }
+    }
+    return dst;
+error:
+    assert(PyErr_Occurred());
+    return AMQP_EMPTY_ARRAY;
+}
+
 
 _PYRMQ_INLINE int64_t RabbitMQ_now_usec(void)
 {
@@ -217,9 +411,7 @@ _PYRMQ_INLINE int RabbitMQ_wait_timeout(int sockfd, double timeout)
 _PYRMQ_INLINE void
 basic_properties_to_PyDict(amqp_basic_properties_t *props, PyObject *p)
 {
-    register PyObject *key = NULL;
     register PyObject *value = NULL;
-    PyObject *h = NULL;
 
     if (props->_flags & AMQP_BASIC_CONTENT_TYPE_FLAG) {
         value = PySTRING_FROM_AMQBYTES(props->content_type);
@@ -269,78 +461,156 @@ basic_properties_to_PyDict(amqp_basic_properties_t *props, PyObject *p)
         value = PyInt_FromLong(props->timestamp);
         PyDICT_SETSTR_DECREF(p, "timestamp", value);
     }
-
-    h = PyDict_New();
-    PyDICT_SETSTR_DECREF(p, "headers", h);
-
     if (props->_flags & AMQP_BASIC_HEADERS_FLAG) {
-        int i;
+        PyDICT_SETSTR_DECREF(p, "headers",
+            AMQTable_toPyDict(&(props->headers)));
+    }
+}
 
-        for (i = 0; i < props->headers.num_entries; ++i) {
-            switch (props->headers.entries[i].value.kind) {
+
+static PyObject*
+AMQTable_toPyDict(amqp_table_t *table)
+{
+    register PyObject *key = NULL;
+    register PyObject *value = NULL;
+    PyObject *dict = NULL;
+    dict = PyDict_New();
+
+    if (table) {
+        int i;
+        for (i = 0; i < table->num_entries; ++i, key=value=NULL) {
+            switch (table->entries[i].value.kind) {
                 case AMQP_FIELD_KIND_BOOLEAN:
-                    key = PySTRING_FROM_AMQBYTES(props->headers.entries[i].key);
-                    value = PyBool_FromLong(AMQTable_HVAL(props, i, boolean));
-                    PyDICT_SETKV_DECREF(h, key, value);
+                    value = PyBool_FromLong(AMQTable_VAL(table, i, boolean));
                     break;
                 case AMQP_FIELD_KIND_I8:
-                    key = PySTRING_FROM_AMQBYTES(props->headers.entries[i].key);
-                    value = PyInt_FromLong(AMQTable_HVAL(props, i, i8));
-                    PyDICT_SETKV_DECREF(h, key, value);
+                    value = PyInt_FromLong(AMQTable_VAL(table, i, i8));
                     break;
                 case AMQP_FIELD_KIND_I16:
-                    key = PySTRING_FROM_AMQBYTES(props->headers.entries[i].key);
-                    value = PyInt_FromLong(AMQTable_HVAL(props, i, i16));
-                    PyDICT_SETKV_DECREF(h, key, value);
+                    value = PyInt_FromLong(AMQTable_VAL(table, i, i16));
                     break;
                 case AMQP_FIELD_KIND_I32:
-                    key = PySTRING_FROM_AMQBYTES(props->headers.entries[i].key);
-                    value = PyInt_FromLong(AMQTable_HVAL(props, i, i32));
-                    PyDICT_SETKV_DECREF(h, key, value);
+                    value = PyInt_FromLong(AMQTable_VAL(table, i, i32));
                     break;
                 case AMQP_FIELD_KIND_I64:
-                    key = PySTRING_FROM_AMQBYTES(props->headers.entries[i].key);
-                    value = PyLong_FromLong(AMQTable_HVAL(props, i, i64));
-                    PyDICT_SETKV_DECREF(h, key, value);
+                    value = PyLong_FromLong(AMQTable_VAL(table, i, i64));
                     break;
                 case AMQP_FIELD_KIND_U8:
-                    key = PySTRING_FROM_AMQBYTES(props->headers.entries[i].key);
-                    value = PyLong_FromUnsignedLong(AMQTable_HVAL(props, i, u8));
-                    PyDICT_SETKV_DECREF(h, key, value);
+                    value = PyLong_FromUnsignedLong(
+                            AMQTable_VAL(table, i, u8));
                     break;
                 case AMQP_FIELD_KIND_U16:
-                    key = PySTRING_FROM_AMQBYTES(props->headers.entries[i].key);
-                    value = PyLong_FromUnsignedLong(AMQTable_HVAL(props, i, u16));
-                    PyDICT_SETKV_DECREF(h, key, value);
+                    value = PyLong_FromUnsignedLong(
+                            AMQTable_VAL(table, i, u16));
                     break;
                 case AMQP_FIELD_KIND_U32:
-                    key = PySTRING_FROM_AMQBYTES(props->headers.entries[i].key);
-                    value = PyLong_FromUnsignedLong(AMQTable_HVAL(props, i, u32));
-                    PyDICT_SETKV_DECREF(h, key, value);
+                    value = PyLong_FromUnsignedLong(
+                            AMQTable_VAL(table, i, u32));
                     break;
                 case AMQP_FIELD_KIND_U64:
-                    key = PySTRING_FROM_AMQBYTES(props->headers.entries[i].key);
-                    value = PyLong_FromUnsignedLong(AMQTable_HVAL(props, i, u64));
-                    PyDICT_SETKV_DECREF(h, key, value);
+                    value = PyLong_FromUnsignedLong(
+                            AMQTable_VAL(table, i, u64));
                     break;
                 case AMQP_FIELD_KIND_F32:
-                    key = PySTRING_FROM_AMQBYTES(props->headers.entries[i].key);
-                    value = PyFloat_FromDouble(AMQTable_HVAL(props, i, f32));
-                    PyDICT_SETKV_DECREF(h, key, value);
+                    value = PyFloat_FromDouble(AMQTable_VAL(table, i, f32));
                     break;
                 case AMQP_FIELD_KIND_F64:
-                    key = PySTRING_FROM_AMQBYTES(props->headers.entries[i].key);
-                    value = PyFloat_FromDouble(AMQTable_HVAL(props, i, f64));
-                    PyDICT_SETKV_DECREF(h, key, value);
+                    value = PyFloat_FromDouble(AMQTable_VAL(table, i, f64));
                     break;
                 case AMQP_FIELD_KIND_UTF8:
-                    key = PySTRING_FROM_AMQBYTES(props->headers.entries[i].key);
-                    value = PySTRING_FROM_AMQBYTES(AMQTable_HVAL(props, i, bytes));
-                    PyDICT_SETKV_DECREF(h, key, value);
+                    value = PySTRING_FROM_AMQBYTES(
+                            AMQTable_VAL(table, i, bytes));
+                    break;
+                case AMQP_FIELD_KIND_TABLE:
+                    value = AMQTable_toPyDict(&(AMQTable_VAL(table, i, table)));
+                    break;
+                case AMQP_FIELD_KIND_ARRAY:
+                    value = AMQArray_toPyList(&(AMQTable_VAL(table, i, array)));
                     break;
             }
+
+            key = AMQTable_TO_PYKEY(table, i);
+            if (value)
+                PyDICT_SETKV_DECREF(dict, key, value);
+            else
+                /* unsupported type */
+                PyDICT_SETNONE_DECREF(dict, key);
+
         }
     }
+    return dict;
+}
+
+
+static PyObject*
+AMQArray_toPyList(amqp_array_t *array)
+{
+    register PyObject *value = NULL;
+    PyObject *list = NULL;
+    list = PyList_New(array->num_entries);
+
+    if (array) {
+        int i;
+        for (i = 0; i < array->num_entries; ++i, value=NULL) {
+            switch (array->entries[i].kind) {
+                case AMQP_FIELD_KIND_BOOLEAN:
+                    value = PyBool_FromLong(AMQArray_VAL(array, i, boolean));
+                    break;
+                case AMQP_FIELD_KIND_I8:
+                    value = PyInt_FromLong(AMQArray_VAL(array, i, i8));
+                    break;
+                case AMQP_FIELD_KIND_I16:
+                    value = PyInt_FromLong(AMQArray_VAL(array, i, i16));
+                    break;
+                case AMQP_FIELD_KIND_I32:
+                    value = PyInt_FromLong(AMQArray_VAL(array, i, i32));
+                    break;
+                case AMQP_FIELD_KIND_I64:
+                    value = PyLong_FromLong(AMQArray_VAL(array, i, i64));
+                    break;
+                case AMQP_FIELD_KIND_U8:
+                    value = PyLong_FromUnsignedLong(AMQArray_VAL(array, i, u8));
+                    break;
+                case AMQP_FIELD_KIND_U16:
+                    value = PyLong_FromUnsignedLong(
+                            AMQArray_VAL(array, i, u16));
+                    break;
+                case AMQP_FIELD_KIND_U32:
+                    value = PyLong_FromUnsignedLong(
+                            AMQArray_VAL(array, i, u32));
+                    break;
+                case AMQP_FIELD_KIND_U64:
+                    value = PyLong_FromUnsignedLong(
+                            AMQArray_VAL(array, i, u64));
+                    break;
+                case AMQP_FIELD_KIND_F32:
+                    value = PyFloat_FromDouble(AMQArray_VAL(array, i, f32));
+                    break;
+                case AMQP_FIELD_KIND_F64:
+                    value = PyFloat_FromDouble(AMQArray_VAL(array, i, f64));
+                    break;
+                case AMQP_FIELD_KIND_UTF8:
+                    value = PySTRING_FROM_AMQBYTES(
+                            AMQArray_VAL(array, i, bytes));
+                    break;
+                case AMQP_FIELD_KIND_TABLE:
+                    value = AMQTable_toPyDict(&(AMQArray_VAL(array, i, table)));
+                    break;
+                case AMQP_FIELD_KIND_ARRAY:
+                    value = AMQArray_toPyList(&(AMQArray_VAL(array, i, array)));
+                    break;
+                default:
+                    /* unsupported type */
+                    Py_INCREF(Py_None);
+                    value = Py_None;
+                    break;
+            }
+
+            PyList_SET_ITEM(list, i, value);
+
+        }
+    }
+    return list;
 }
 
 
@@ -473,27 +743,27 @@ PyRabbitMQ_HandlePollError(int ready)
 }
 
 
-int PyRabbitMQ_HandleAMQError(amqp_rpc_reply_t reply,
-                              PyObject *exc_cls, const char *context)
+int PyRabbitMQ_HandleAMQError(PyRabbitMQ_Connection *self, unsigned int channel,
+        amqp_rpc_reply_t reply, const char *context)
 {
     char errorstr[1024];
 
     switch (reply.reply_type) {
         case AMQP_RESPONSE_NORMAL:
-            return 1;
+            return 0;
 
         case AMQP_RESPONSE_NONE:
             snprintf(errorstr, sizeof(errorstr),
                     "%s: missing RPC reply type!", context);
-            break;
+            goto connerror;
 
         case AMQP_RESPONSE_LIBRARY_EXCEPTION:
             snprintf(errorstr, sizeof(errorstr), "%s: %s",
                     context,
                     reply.library_error
-                        ? strerror(reply.library_error)
+                        ? amqp_error_string(reply.library_error)
                         : "(end-of-stream)");
-            break;
+            goto connerror;
 
         case AMQP_RESPONSE_SERVER_EXCEPTION:
 
@@ -505,7 +775,7 @@ int PyRabbitMQ_HandleAMQError(amqp_rpc_reply_t reply,
                             context,
                             m->reply_code,
                             (int) m->reply_text.len, (char *) m->reply_text.bytes);
-                    break;
+                    goto connerror;
                 }
                 case AMQP_CHANNEL_CLOSE_METHOD: {
                     amqp_channel_close_t *m = (amqp_channel_close_t *) reply.reply.decoded;
@@ -513,18 +783,24 @@ int PyRabbitMQ_HandleAMQError(amqp_rpc_reply_t reply,
                         "%s: server channel error %d, message: %.*s",
                             context, m->reply_code,
                             (int) m->reply_text.len, (char *) m->reply_text.bytes);
-                    break;
+                    goto chanerror;
                 }
                 default:
                     snprintf(errorstr, sizeof(errorstr),
                         "%s: unknown server error, method id 0x%08X",
                             context, reply.reply.id);
-                    break;
+                    goto connerror;
             }
             break;
     }
-    PyErr_SetString(exc_cls, errorstr);
-    return 0;
+connerror:
+    PyErr_SetString(PyRabbitMQExc_ConnectionError, errorstr);
+    PyRabbitMQ_Connection_close(self);
+    return PYRABBITMQ_CONNECTION_ERROR;
+chanerror:
+    PyErr_SetString(PyRabbitMQExc_ChannelError, errorstr);
+    PyRabbitMQ_revive_channel(self, channel);
+    return PYRABBITMQ_CHANNEL_ERROR;
 }
 
 
@@ -536,6 +812,21 @@ void PyRabbitMQ_SetErr_UnexpectedHeader(amqp_frame_t* frame)
     PyErr_SetString(PyRabbitMQExc_ChannelError, errorstr);
 }
 
+unsigned int
+PyRabbitMQ_revive_channel(PyRabbitMQ_Connection *self, unsigned int channel)
+{
+    int status = -1;
+    amqp_channel_close_ok_t req;
+    status = amqp_send_method(self->conn, (amqp_channel_t)channel,
+                            AMQP_CHANNEL_CLOSE_OK_METHOD, &req);
+    if (status < 0) {
+        PyErr_SetString(PyRabbitMQExc_ConnectionError, "Couldn't revive channel");
+        PyRabbitMQ_Connection_close(self);
+        return 1;
+    }
+    return PyRabbitMQ_Connection_create_channel(self, channel);
+}
+
 /* ------: Connection :--------------------------------------------------- */
 
 /*
@@ -661,9 +952,8 @@ PyRabbitMQ_Connection_connect(PyRabbitMQ_Connection *self)
                        self->frame_max, self->heartbeat,
                        AMQP_SASL_METHOD_PLAIN, self->userid, self->password);
     Py_END_ALLOW_THREADS;
-    if (!PyRabbitMQ_HandleAMQError(reply,
-            PyRabbitMQExc_ConnectionError, "Couldn't log in"))
-        goto error;
+    if (PyRabbitMQ_HandleAMQError(self, 0, reply, "Couldn't log in"))
+        goto bail;
 
     /* after tune */
     self->connected = 1;
@@ -700,6 +990,20 @@ PyRabbitMQ_Connection_close(PyRabbitMQ_Connection *self)
 }
 
 
+unsigned int
+PyRabbitMQ_Connection_create_channel(PyRabbitMQ_Connection *self, unsigned int channel)
+{
+    amqp_rpc_reply_t reply;
+
+    Py_BEGIN_ALLOW_THREADS;
+    amqp_channel_open(self->conn, channel);
+    reply = amqp_get_rpc_reply(self->conn);
+    Py_END_ALLOW_THREADS;
+
+    return PyRabbitMQ_HandleAMQError(self, 0, reply, "Couldn't create channel");
+}
+
+
 /*
  * Connection._channel_open
  */
@@ -707,7 +1011,6 @@ static PyObject *
 PyRabbitMQ_Connection_channel_open(PyRabbitMQ_Connection *self, PyObject *args)
 {
     unsigned int channel;
-    amqp_rpc_reply_t reply;
 
     if (PyRabbitMQ_Not_Connected(self))
         goto bail;
@@ -715,19 +1018,10 @@ PyRabbitMQ_Connection_channel_open(PyRabbitMQ_Connection *self, PyObject *args)
     if (!PyArg_ParseTuple(args, "I", &channel))
         goto bail;
 
-    Py_BEGIN_ALLOW_THREADS;
-    amqp_channel_open(self->conn, channel);
-    reply = amqp_get_rpc_reply(self->conn);
-    Py_END_ALLOW_THREADS;
-
-    if (!PyRabbitMQ_HandleAMQError(reply,
-            PyRabbitMQExc_ChannelError, "Couldn't create channel"))
-        goto error;
+    if (PyRabbitMQ_Connection_create_channel(self, channel))
+        goto bail;;
 
     Py_RETURN_NONE;
-
-error:
-    PyRabbitMQ_Connection_close(self);
 bail:
     return 0;
 }
@@ -736,12 +1030,25 @@ bail:
 /*
  * Connection._channel_close
  */
+
+unsigned int
+PyRabbitMQ_Connection_destroy_channel(PyRabbitMQ_Connection *self,
+                                      unsigned int channel)
+{
+    amqp_rpc_reply_t reply;
+    Py_BEGIN_ALLOW_THREADS;
+    reply = amqp_channel_close(self->conn, channel, AMQP_REPLY_SUCCESS);
+    amqp_maybe_release_buffers(self->conn);
+    Py_END_ALLOW_THREADS;
+
+    return PyRabbitMQ_HandleAMQError(self, channel, reply, "Couldn't close channel");
+}
+
 static PyObject*
 PyRabbitMQ_Connection_channel_close(PyRabbitMQ_Connection *self,
                                     PyObject *args)
 {
     unsigned int channel = 0;
-    amqp_rpc_reply_t reply;
 
     if (PyRabbitMQ_Not_Connected(self))
         goto error;
@@ -749,13 +1056,7 @@ PyRabbitMQ_Connection_channel_close(PyRabbitMQ_Connection *self,
     if (!PyArg_ParseTuple(args, "I", &channel))
         goto error;
 
-    Py_BEGIN_ALLOW_THREADS;
-    reply = amqp_channel_close(self->conn, channel, AMQP_REPLY_SUCCESS);
-    amqp_maybe_release_buffers(self->conn);
-    Py_END_ALLOW_THREADS;
-
-    if (!PyRabbitMQ_HandleAMQError(reply,
-            PyRabbitMQExc_ChannelError, "Couldn't close channel"))
+    if (PyRabbitMQ_Connection_destroy_channel(self, channel))
         goto error;
 
     Py_RETURN_NONE;
@@ -1001,13 +1302,10 @@ PyRabbitMQ_Connection_queue_bind(PyRabbitMQ_Connection *self,
     amqp_maybe_release_buffers(self->conn);
     Py_END_ALLOW_THREADS;
 
-    if (!PyRabbitMQ_HandleAMQError(reply,
-            PyRabbitMQExc_ChannelError, "queue.bind"))
-        goto error;
+    if (PyRabbitMQ_HandleAMQError(self, channel, reply, "queue.bind"))
+        goto bail;
 
     Py_RETURN_NONE;
-error:
-    PyRabbitMQ_Connection_close(self);
 bail:
     return 0;
 }
@@ -1052,13 +1350,10 @@ PyRabbitMQ_Connection_queue_unbind(PyRabbitMQ_Connection *self,
     amqp_maybe_release_buffers(self->conn);
     Py_END_ALLOW_THREADS;
 
-    if (!PyRabbitMQ_HandleAMQError(reply,
-            PyRabbitMQExc_ChannelError, "queue.unbind"))
-        goto error;
+    if (PyRabbitMQ_HandleAMQError(self, channel, reply, "queue.unbind"))
+        goto bail;
 
     Py_RETURN_NONE;
-error:
-    PyRabbitMQ_Connection_close(self);
 bail:
     return 0;
 }
@@ -1096,13 +1391,11 @@ PyRabbitMQ_Connection_queue_delete(PyRabbitMQ_Connection *self,
     amqp_maybe_release_buffers(self->conn);
     Py_END_ALLOW_THREADS;
 
-    if (ok == NULL && !PyRabbitMQ_HandleAMQError(reply,
-            PyRabbitMQExc_ChannelError, "queue.delete"))
-        goto error;
+    if (ok == NULL && PyRabbitMQ_HandleAMQError(self, channel,
+            reply, "queue.delete"))
+        goto bail;
 
     return PyInt_FromLong((long)ok->message_count);
-error:
-    PyRabbitMQ_Connection_close(self);
 bail:
     return 0;
 }
@@ -1152,9 +1445,8 @@ PyRabbitMQ_Connection_queue_declare(PyRabbitMQ_Connection *self,
     reply = amqp_get_rpc_reply(self->conn);
     Py_END_ALLOW_THREADS;
 
-    if (!PyRabbitMQ_HandleAMQError(reply,
-            PyRabbitMQExc_ChannelError, "queue.declare"))
-        goto error;
+    if (PyRabbitMQ_HandleAMQError(self, channel, reply, "queue.declare"))
+        goto bail;
 
     if ((ret = PyTuple_New(3)) == NULL) goto bail;
     PyTuple_SET_ITEM(ret, 0, PyString_FromStringAndSize(ok->queue.bytes,
@@ -1162,8 +1454,6 @@ PyRabbitMQ_Connection_queue_declare(PyRabbitMQ_Connection *self,
     PyTuple_SET_ITEM(ret, 1, PyInt_FromLong((long)ok->message_count));
     PyTuple_SET_ITEM(ret, 2, PyInt_FromLong((long)ok->consumer_count));
     return ret;
-error:
-    PyRabbitMQ_Connection_close(self);
 bail:
     return 0;
 }
@@ -1197,13 +1487,10 @@ PyRabbitMQ_Connection_queue_purge(PyRabbitMQ_Connection *self,
     amqp_maybe_release_buffers(self->conn);
     Py_END_ALLOW_THREADS;
 
-    if (!PyRabbitMQ_HandleAMQError(reply,
-            PyRabbitMQExc_ChannelError, "queue.purge"))
-        goto error;
+    if (PyRabbitMQ_HandleAMQError(self, channel, reply, "queue.purge"))
+        goto bail;
 
     return PyInt_FromLong((long)ok->message_count);
-error:
-    PyRabbitMQ_Connection_close(self);
 bail:
     return 0;
 }
@@ -1253,12 +1540,9 @@ PyRabbitMQ_Connection_exchange_declare(PyRabbitMQ_Connection *self,
     reply = amqp_get_rpc_reply(self->conn);
     Py_END_ALLOW_THREADS;
 
-    if (!PyRabbitMQ_HandleAMQError(reply,
-            PyRabbitMQExc_ChannelError, "exchange.declare"))
-        goto error;
+    if (PyRabbitMQ_HandleAMQError(self, channel, reply, "exchange.declare"))
+        goto bail;
     Py_RETURN_NONE;
-error:
-    PyRabbitMQ_Connection_close(self);
 bail:
     return 0;
 }
@@ -1291,13 +1575,10 @@ PyRabbitMQ_Connection_exchange_delete(PyRabbitMQ_Connection *self,
     amqp_maybe_release_buffers(self->conn);
     Py_END_ALLOW_THREADS;
 
-    if (!PyRabbitMQ_HandleAMQError(reply,
-            PyRabbitMQExc_ChannelError, "exchange.delete"))
-        goto error;
+    if (PyRabbitMQ_HandleAMQError(self, channel, reply, "exchange.delete"))
+        goto bail;
 
     Py_RETURN_NONE;
-error:
-    PyRabbitMQ_Connection_close(self);
 bail:
     return 0;
 }
@@ -1357,7 +1638,7 @@ PyRabbitMQ_Connection_basic_publish(PyRabbitMQ_Connection *self,
     Py_RETURN_NONE;
 
 error:
-    PyRabbitMQ_Connection_close(self);
+    PyRabbitMQ_revive_channel(self, channel);
 bail:
     return 0;
 }
@@ -1392,7 +1673,7 @@ PyRabbitMQ_Connection_basic_ack(PyRabbitMQ_Connection *self,
 
     Py_RETURN_NONE;
 error:
-    PyRabbitMQ_Connection_close(self);
+    PyRabbitMQ_revive_channel(self, channel);
 bail:
     return 0;
 }
@@ -1426,7 +1707,7 @@ static PyObject *PyRabbitMQ_Connection_basic_reject(PyRabbitMQ_Connection *self,
 
     Py_RETURN_NONE;
 error:
-    PyRabbitMQ_Connection_close(self);
+    PyRabbitMQ_revive_channel(self, channel);
 bail:
     return 0;
 }
@@ -1446,7 +1727,7 @@ PyRabbitMQ_Connection_basic_cancel(PyRabbitMQ_Connection *self,
     amqp_rpc_reply_t reply;
 
     if (PyRabbitMQ_Not_Connected(self))
-        goto bail;
+        goto ok;
 
     if (!PyArg_ParseTuple(args, "IO", &channel, &consumer_tag))
         goto bail;
@@ -1459,13 +1740,11 @@ PyRabbitMQ_Connection_basic_cancel(PyRabbitMQ_Connection *self,
     amqp_maybe_release_buffers(self->conn);
     Py_END_ALLOW_THREADS;
 
-    if (!PyRabbitMQ_HandleAMQError(reply,
-            PyRabbitMQExc_ChannelError, "basic.cancel"))
-        goto error;
+    if (PyRabbitMQ_HandleAMQError(self, channel, reply, "basic.cancel"))
+        goto bail;
 
+ok:
     Py_RETURN_NONE;
-error:
-    PyRabbitMQ_Connection_close(self);
 bail:
     return 0;
 }
@@ -1514,13 +1793,10 @@ PyRabbitMQ_Connection_basic_consume(PyRabbitMQ_Connection *self,
     reply = amqp_get_rpc_reply(self->conn);
     Py_END_ALLOW_THREADS;
 
-    if (!PyRabbitMQ_HandleAMQError(reply,
-            PyRabbitMQExc_ChannelError, "basic.consume"))
-        goto error;
+    if (PyRabbitMQ_HandleAMQError(self, channel, reply, "basic.consume"))
+        goto bail;
 
     return PySTRING_FROM_AMQBYTES(ok->consumer_tag);
-error:
-    PyRabbitMQ_Connection_close(self);
 bail:
     return 0;
 }
@@ -1581,13 +1857,10 @@ PyRabbitMQ_Connection_flow(PyRabbitMQ_Connection *self,
     amqp_maybe_release_buffers(self->conn);
     Py_END_ALLOW_THREADS;
 
-    if (!PyRabbitMQ_HandleAMQError(reply,
-            PyRabbitMQExc_ChannelError, "channel.flow"))
-        goto error;
+    if (PyRabbitMQ_HandleAMQError(self, channel, reply, "channel.flow"))
+        goto bail;
 
     Py_RETURN_NONE;
-error:
-    PyRabbitMQ_Connection_close(self);
 bail:
     return 0;
 }
@@ -1617,13 +1890,10 @@ PyRabbitMQ_Connection_basic_recover(PyRabbitMQ_Connection *self,
     amqp_maybe_release_buffers(self->conn);
     Py_END_ALLOW_THREADS;
 
-    if (!PyRabbitMQ_HandleAMQError(reply,
-            PyRabbitMQExc_ChannelError, "basic.recover"))
-        goto error;
+    if (PyRabbitMQ_HandleAMQError(self, channel, reply, "basic.recover"))
+        goto bail;
 
     Py_RETURN_NONE;
-error:
-    PyRabbitMQ_Connection_close(self);
 bail:
     return 0;
 }
@@ -1705,9 +1975,8 @@ PyRabbitMQ_Connection_basic_get(PyRabbitMQ_Connection *self,
                            (amqp_boolean_t)no_ack);
     Py_END_ALLOW_THREADS;
 
-    if (!PyRabbitMQ_HandleAMQError(reply,
-            PyRabbitMQExc_ChannelError, "basic.get"))
-        goto error;
+    if (PyRabbitMQ_HandleAMQError(self, channel, reply, "basic.get"))
+        goto bail;
     if (reply.reply.id != AMQP_BASIC_GET_OK_METHOD)
         goto empty;
 
@@ -1733,7 +2002,6 @@ PyRabbitMQ_Connection_basic_get(PyRabbitMQ_Connection *self,
         }
     }
     return p;
-
 error:
     PyRabbitMQ_Connection_close(self);
 bail:
