@@ -15,7 +15,7 @@ ConnectionError = _librabbitmq.ConnectionError
 ChannelError = _librabbitmq.ChannelError
 
 
-__version__ = '0.9.9'
+__version__ = '1.0.0'
 __all__ = ['Connection', 'Message', 'ConnectionError', 'ChannelError']
 
 
@@ -33,6 +33,7 @@ class Message(object):
     def reject(self):
         return self.channel.basic_reject(self.delivery_info['delivery_tag'])
 
+
 class Channel(object):
     Message = Message
     is_open = False
@@ -43,6 +44,12 @@ class Channel(object):
         self.next_consumer_tag = itertools.count(1).next
         self.no_ack_consumers = set()
 
+    def __enter__(self):
+        return self
+
+    def __exit__(self, *exc_info):
+        self.close()
+
     def basic_qos(self, prefetch_size=0, prefetch_count=0, _global=False):
         return self.connection._basic_qos(self.channel_id,
                 prefetch_size, prefetch_count, _global)
@@ -132,7 +139,9 @@ class Channel(object):
         return self.connection._queue_unbind(self.channel_id,
                 queue, exchange, binding_key, arguments or {})
 
-    def queue_delete(self, queue='', if_unused=False, if_empty=False):
+    def queue_delete(self, queue='', if_unused=False, if_empty=False,
+            nowait=False):
+        """nowait argument is not supported."""
         return self.connection._queue_delete(self.channel_id,
                 queue, if_unused, if_empty)
 
@@ -167,6 +176,12 @@ class Connection(_librabbitmq.Connection):
         if not lazy:
             self.connect()
 
+    def __enter__(self):
+        return self
+
+    def __exit__(self, *exc_info):
+        self.close()
+
     def reconnect(self):
         self.close()
         self.connect()
